package logwire.web.bo.field.parse;

import logwire.core.bo.annotation.Choice;
import logwire.core.bo.annotation.ForeignKey;
import logwire.core.bo.field.BizArray;
import logwire.core.bo.field.BizMany;
import logwire.core.bo.field.BizOne;
import logwire.core.exceptions.ApplicationException;

import java.lang.reflect.Field;
import java.lang.reflect.Type;
import java.math.BigDecimal;
import java.time.OffsetDateTime;

public abstract class FieldParserUtil {

    private FieldParserUtil() {
    }

    /**
     * 获取解析注解的 parser
     *
     * @param field
     * @return
     */
    public static FieldParser getParser(Field field) {
        if (field.getAnnotation(Choice.class) != null) {
            if (field.getType().equals(String.class)) {
                return ChoiceFieldParser.getInstance();
            } else if (field.getType().equals(BizArray.class)) {
                return BizArrayFieldParser.getInstance();
            } else {
                throw new ApplicationException("@Choice 只能用与 String/BizArray 类型");
            }
        } else if (field.getAnnotation(ForeignKey.class) != null) {
            //弱关联
            if (field.getType().equals(BizOne.class)||field.getType().equals(BizMany.class)) {
                return ForeignKeyFieldParser.getInstance();
            } else {
                throw new ApplicationException("@ForeignKey 只能用与 BizOne/BizMany 类型");
            }

        }
        return getParserByFieldType(field.getType());
    }

    public static FieldParser getParserByFieldType(Type clazz) {
        // 此时一定是存在 Field 类型的 annotation
        // 根据字段类型来推断 Parser
        if (clazz.equals(String.class)) {
            return StringFieldParser.getInstance();
        } else if (clazz.equals(Integer.class) || clazz.equals(int.class)) {
            return IntegerFieldParser.getInstance();
        } else if (clazz.equals(Short.class) || clazz.equals(short.class)) {
            return SmallIntegerFieldParser.getInstance();
        } else if (clazz.equals(Long.class) || clazz.equals(long.class)) {
            return BigIntegerFieldParser.getInstance();
        } else if (clazz.equals(OffsetDateTime.class) || clazz.equals(java.util.Date.class)) {
            return DateTimeFieldParser.getInstance();
        } else if (clazz.equals(Boolean.class) || clazz.equals(boolean.class)) {
            return BooleanFieldParser.getInstance();
        } else if (clazz.equals(Double.class) || clazz.equals(double.class) || clazz.equals(float.class)||clazz.equals(Float.class)||clazz.equals(BigDecimal.class)) {
            return DecimalFieldParser.getInstance();
        } else if (clazz.equals(BizArray.class)) {
            return BizArrayFieldParser.getInstance();
        } else if (clazz.equals(BizOne.class)||clazz.equals(BizMany.class)) {
            return ForeignKeyFieldParser.getInstance();
        }
        throw new ApplicationException("该字段类型无法识别");
    }
}
