package logwire.web.bo.field.parse;

import logwire.core.bo.annotation.ForeignKey;
import logwire.core.bo.field.BizOne;
import logwire.core.exceptions.ApplicationException;
import logwire.core.meta.model.fields.ForeignKeyField;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;

public class ForeignKeyFieldParser extends FieldParser<ForeignKeyField> {
    @Override
    protected ForeignKeyField newField(String name, Field field) {
        /**
         *
         * BizOne类型表示外键字段，
         * TODO 若使用了@ForeignKey 弱关联
         *
         * */
        if (field.getType().equals(BizOne.class)) {
            if (field.getAnnotation(ForeignKey.class) != null) {
                return new ForeignKeyField(name, field.getAnnotation(ForeignKey.class).referBoName());
            } else {
                Type genericType = field.getGenericType();
                ParameterizedType parameterizedType = (ParameterizedType) genericType;
                Type actualTypeArgument = parameterizedType.getActualTypeArguments()[0];
                return new ForeignKeyField(name, actualTypeArgument.getTypeName());
            }
        } else {
            throw new ApplicationException("非BizOne类型不能作为外键");
        }
    }

    private ForeignKeyFieldParser() {
    }

    public static ForeignKeyFieldParser getInstance() {
        return ForeignKeyFieldParser.SingletonHolder.instance;
    }

    private static class SingletonHolder {
        private static ForeignKeyFieldParser instance = new ForeignKeyFieldParser();
    }

}
