package logwire.web.bo.proxy;

import com.google.common.collect.Lists;
import logwire.core.bo.handler.ObjectOperationHandler;
import logwire.core.bo.object.BizObject;
import logwire.web.bo.BoRelationDefinition;
import logwire.web.bo.provider.DefaultTypeOperationProvider;
import net.bytebuddy.ByteBuddy;
import net.bytebuddy.implementation.MethodDelegation;
import net.bytebuddy.matcher.ElementMatchers;

import java.util.List;
import java.util.Map;

public class Proxy {
    /**
     * 初始化 BO的 service : TypeOperationProvider
     *
     * @param boRelationDefinition
     * @return
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws NoSuchFieldException
     */
    public static Object createTypeOperationProvider(BoRelationDefinition boRelationDefinition) throws IllegalAccessException, InstantiationException, NoSuchFieldException {
        Class<? extends BizObject> boClass = boRelationDefinition.getBoClass();

        // 动态生成service
        // 若直接继承了BizObject,则动态代理继承DefaultTypeOperationProvider，
        // 否则继承父类中的service对应的TypeOperationProvider
        Class superClass = DefaultTypeOperationProvider.class;
        if (boClass.getSuperclass() != BizObject.class) {
            // parent TypeOperationProvider
            superClass = boClass.getSuperclass().getField("service").getClass();
        }
        return new ByteBuddy().subclass(superClass)
                .implement(boRelationDefinition.getTypeOperationProviders())
                .method(ElementMatchers.any())
                .intercept(MethodDelegation.to(new TypeOperationInterceptor(boRelationDefinition)))
                .make()
                .load(boClass.getClassLoader())
                .getLoaded().newInstance();
    }

    public static Object createObjectOperationProvider(BoRelationDefinition boRelationDefinition, Map<String, BoRelationDefinition> boRelationDefinitions) throws IllegalAccessException, InstantiationException {
        Class<? extends BizObject> boClass = boRelationDefinition.getBoClass();
        Class boSuperClass = boClass.getSuperclass();
        List<ObjectOperationHandler> interfaces = Lists.newArrayList(boRelationDefinition.getObjectOperationHandlers());
        while (boSuperClass != BizObject.class) {
            interfaces.addAll(boRelationDefinitions.get(boSuperClass.getName()).getObjectOperationHandlers());
            boSuperClass = boSuperClass.getSuperclass();
        }
        return new ByteBuddy().subclass(boClass.getSuperclass())
                .implement(boRelationDefinition.getObjectOperationProviders())
                .method(ElementMatchers.any())
                .intercept(MethodDelegation.to(new ObjectOperationInterceptor(boRelationDefinition)))
                .make()
                .load(boClass.getClassLoader())
                .getLoaded().newInstance();


    }
}
